#pragma rtGlobals=3		// Use modern global access method.
#pragma TextEncoding 	= "UTF-8"
#pragma IgorVersion		= 8.0
#pragma ModuleName		= SpectraWatchDog
#pragma version			= 1.02
#pragma DefaultTab		= {3,20,4}			// Set default tab width in Igor Pro 9 and later

static Constant kVersion = 1.02
static StrConstant kVersionDate = "07/2023"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 22182																// the project node on IgorExchange
static StrConstant ksShortTitle = "Spectra WatchDog"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//	
//	Spectra Folder Watchdog
//	
//	Writes text logs of spectra metadata from files in a designated folder. Logs can be complemented
//	with user information, i.e., the log is updated with values entered in a panel. Folders can either
//	be read once or watched continuously for new spectra files.
//	
//	Provided with support for the following file formats:
//		- ibw:	Scienta SES photoemission spectra
//		- zip:	Zip files containing spectra (default content: *.ibw)
//
// 2023-06-20 - ver. 1.01:	Added support for zipped IBW files (Igor 9).
// 2023-07-28 - ver. 1.02:	Initial public release.
//							Fixed bug: Entries we note updated correctly for duplicate file names within zip files.
//							Added support for extension lists.
//							Generalized info function support.
//________________________________________________________________________________________________

static Constant kCheckFolderPeriod		= 60													// In "ticks" which are ~1/60th of a second.
static StrConstant kLogFileEnding		= ".log"												// writes text files with this ending
static StrConstant kDelimiterChar		= "\t"													// type of delimiter for entries in the log file
static StrConstant kSettingsFileName	= "Spectra WatchDog settings.dat"
static StrConstant kWatchdogWorkingDir 	= "Spectra_Watchdog"

static StrConstant kDefaultColumnWidhts = "2,2.5,1,1,1,1,1,1,1,1,1,1,2.5,2.5,1.75,1.25,1.75,1.25,1.25"		// default width of the columns in the displayed table
static StrConstant kDefaultColumnColors = "0,1,0,0,0,0,0,0,0,1,4,4,8,8,0,0,0,0,0"				// color columns in predefined colors (0-12)
static StrConstant kDefaultRepeatColumn = "0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0"				// which columns repeat their entries each time (0 => no repeat, 1 => repeat)
static StrConstant kDefaultExtensionStr	= "ibw"													// looks for this file format

Menu "Spectra Tools", dynamic
	"-"
	"Folder Watchdog...",	/Q, SpectraWatchDog#buildWatchDogPanel()
	"Folder Quick-log...",	/Q, SpectraWatchDog#QuickLogFromFolder()
End

//################################################################################################

function/wave WDfunc_IBW_SESInfo(string filePath)												// ibw:	Scienta SES photoemission spectra
	string infoStr = ""
	DFREF savDF = GetDataFolderDFR()
	SetDataFolder NewFreeDataFolder()															// set temp folder to load data wave here	
		LoadWave/O/Q filePath
		if (V_flag > 0)
			wave data = $StringFromList(0,S_WaveNames);	 infoStr = note(data)
		endif
	SetDataFolder savDF
	
	if (!strlen(infoStr) || !StringMatch(infoStr[0,4],"[SES]"))									// look in the notes for the SES header
		return $""
	endif
	// ++++++++++++++++++++++++++++++++++++++++
	GetFileFolderInfo/Q/Z filePath																// gets file creation time from the OS
	string LoadPath  = GetInfo("LoadPath", infoStr)
	string FullPath  = GetInfo("File", infoStr)
	string SweepNum  = GetInfo("Number of Sweeps", infoStr)
	string StartDate = GetInfo("Date", infoStr)
	string StartTime = GetInfo("Time", infoStr)
	string WriteDate = Secs2Date(V_creationDate,-2)
	string WriteTime = Secs2Time(V_creationDate, 3)
	
	int zDim = DimSize(data,2)
	string TagName = ParseFilePath(3, FullPath, SelectString(strsearch(FullPath,"\\",0) > -1,":","\\"), 0, 0)
	if (!strlen(TagName))
		TagName = ParseFilePath(3, filePath, SelectString(strsearch(FullPath,"\\",0) > -1,":","\\"), 0, 0)
		TagName = RemoveEnding(TagName,GetInfo("Spectrum Name", infoStr))
	endif
	if (StringMatch(GetInfo("Name", infoStr), "Add Dimension"))
		SweepNum = SelectString(str2num(GetInfo("Number of Slices", infoStr)) > 1, num2str(DimSize(data,1)), num2str(zDim == 0 ? 1 : zDim))
	endif
	
	variable year, month, day, hh, mm, ss, time1, time2											// calculate the time the measurement took
	sscanf StartDate, "%d%*[-]%d%*[-]%d", year, month, day
	sscanf StartTime, "%d%*[:]%d%*[:]%d", hh, mm, ss
	time1 = Date2Secs(year, month, day) + hh*60*60 + mm*60 + ss
	sscanf WriteDate, "%d%*[-]%d%*[-]%d", year, month, day
	sscanf WriteTime, "%d%*[:]%d%*[:]%d", hh, mm, ss
	time2 = Date2Secs(year, month, day) + hh*60*60 + mm*60 + ss
	
	Make/Free/T/N=(0,2) info
	info[DimSize(info,0)] = {{"Tag"}			,{TagName}}
	info[DimSize(info,0)] = {{"Sample"}			,{GetInfo("Sample", infoStr)}}
	info[DimSize(info,0)] = {{"E_Pass [eV]"}	,{GetInfo("Pass Energy", infoStr)}}
	info[DimSize(info,0)] = {{"E_Mode"}			,{SelectString(StringMatch(GetInfo("Energy Unit", infoStr), "Kinetic"), "BE", "KE")}}
	info[DimSize(info,0)] = {{"E_Start [eV]"}	,{GetInfo("Low Energy", infoStr)}}
	info[DimSize(info,0)] = {{"E_End [eV]"}		,{GetInfo("High Energy", infoStr)}}
	info[DimSize(info,0)] = {{"E_Step [eV]"}	,{GetInfo("Energy Step", infoStr)}}
	info[DimSize(info,0)] = {{"Dwell Time [ms]"},{GetInfo("Step Time", infoStr)}}
	info[DimSize(info,0)] = {{"# Sweeps"}		,{SweepNum}}
	info[DimSize(info,0)] = {{"E_hv [eV]"}		,{GetInfo("Excitation Energy", infoStr)}}
	info[DimSize(info,0)] = {{"Det. Slit"}		,{""}}
	info[DimSize(info,0)] = {{"Bias [V]"}		,{""}}
	info[DimSize(info,0)] = {{"Purpose"}		,{""}}
	info[DimSize(info,0)] = {{"Comments"}		,{GetInfo("Comments", infoStr)}}
	info[DimSize(info,0)] = {{"Start Date"}		,{StartDate}}
	info[DimSize(info,0)] = {{"Start Time"}		,{StartTime}}
	info[DimSize(info,0)] = {{"Write Date"}		,{WriteDate}}
	info[DimSize(info,0)] = {{"Write Time"}		,{WriteTime}}
	info[DimSize(info,0)] = {{"Elapsed Time"}	,{Secs2Time(time2-time1,3)}}
	return info
end

static function/S GetInfo(string which, string infoStr)
	return StringByKey(which, infoStr, "=", "\r")
end

//################################################################################################

function QuickLogFromFolder()																	// saves a log of all data files in the same folder as the data
	CtrlNamedBackground SpectraWatchDogTask, status
	
	if (NumberByKey("RUN", S_info))
		Print "Caution: Cannot write log while WatchDog is running."
	endif
	
	int i, fileID
	string extList = "", curExt
	string allFuncs = SortList(FunctionList("WDfunc_*",";","KIND:2,VALTYPE:8,PARAM_0_TYPE:8192"), ";", 4)
	for (i=0; i < ItemsInList(allFuncs); i++)
		SplitString/E=("WDfunc_([[:alpha:]]+)_") StringFromList(i, allFuncs), curExt
		if (FindListItem("*."+curExt, extList, ",") < 0)
			extList += "*." + LowerStr(curExt) + ","
		endif
	endfor
	extList = RemoveEnding(extList,",") + SelectString(IgorVersion()>=9,"",",*.zip")			// zip support only from Igor 9
	
	Open/D/R/F=("Data Files ("+extList+"):"+ReplaceString("*",extList,"")+";")/M="Select any data file from folder to write the log file." fileID
	if (strlen(S_Filename) > 0)
		string loadFolder = ParseFilePath(1, S_Filename, ":", 1, 0)
		string folderName = ParseFilePath(0, S_Filename, ":", 1, 1)
		NewPath/O/Q/Z WD_WatchFolderPath, loadFolder
		DeleteFile/Z loadFolder+folderName+kLogFileEnding
		string loadList = IndexedFile(WD_WatchFolderPath,-1,"."+ParseFilePath(4, S_Filename, ":", 0, 0))
		for (i=0; i<ItemsInList(loadList); i+=1)
			SpectraWatchDog#WD_processFile(loadFolder+StringFromList(i, loadList), logFile=loadFolder+folderName+kLogFileEnding)
		endfor
	endif
end

//################################################################################################
//	core functions
//################################################################################################

static function/wave WD_readSpectraInfo(string filePath)										// returns a two-column text wave with all parameters from the input file
	int i
	string allFuncs = SortList(FunctionList("WDfunc_*",";","KIND:2,VALTYPE:8,PARAM_0_TYPE:8192"), ";", 4)
	string funcList = "", curExt, curName, requestedExt = ParseFilePath(4,filePath,":",0,0)
	for (i=0; i < ItemsInList(allFuncs); i++)
		SplitString/E=("WDfunc_([[:alpha:]]+)_(.*)") StringFromList(i, allFuncs), curExt, curName
		if (!CmpStr(curExt,requestedExt))
			funcList += StringFromList(i, allFuncs)
		endif
	endfor
	// ++++++++++++++++++++++++++++++++++++++++
	wave/T/Z info = $""																			// initialize wave reference
	for (i=0; i < ItemsInList(funcList); i++)
		FUNCREF WDdummy_InfoFunc infoParse=$StringFromList(i, funcList)
		wave/T/Z info = infoParse(filePath)
		if (WaveExists(info))																	// valid output
			break
		endif
	endfor
	return info																					// format: col[0] = table column titles, col[1] = table entries
end

function/wave WDdummy_InfoFunc(string filePath)
	return $""
end

//################################################################################################

static function WD_processFile(string filePath [,string logFile, string extList])
	string fileExtension = ParseFilePath(4, filePath, ":", 0, 0)
	string folderPath = "", fileList = filePath, zipNamePrefix = ""								// is the full path already
	int i, error = 0
	
	if (!Cmpstr(fileExtension,"zip"))															// added zip support
		folderPath = SpecialDirPath("Temporary",0,0,0)
		error = IgorVersion() < 9.0
	#if(IgorVersion() >= 9)																		// unzip is supported from Igor 9
		unzipFile/O/Z filePath, RemoveEnding(folderPath,":")
		error = V_flag
	#endif
		if (error)
			Print "Error: Could not unzip file. Zipped files not supported."
			return -3
		endif
		
		string extInsideZip = "ibw"																// fallback is always ibw
		if (!ParamIsDefault(extList))
			extList = RemoveFromList("zip",extList)
			extInsideZip = SelectString(strlen(extList),"ibw",extList)
		endif
		NewPath/O/Q tmpPath, folderPath;	fileList = ""
		for (i=0; i<ItemsInList(extInsideZip); i++)												// grab valid files from extraction location
			fileList += IndexedFile(tmpPath, -1, "."+StringFromList(i,extInsideZip))
		endfor
		
		if (!ItemsInList(fileList))
			Print "Caution: Zipped file \"" + ParseFilePath(0, filePath, ":", 1, 0)+ "\" did not contain any spectra. Skipping ..."
			return -4
		endif
		sPrintf zipNamePrefix, "[%s] ", ParseFilePath(0, filePath, ":", 1, 0)
		KillPath/Z tmpPath
	endif
	
	// +++++++++++++++++++++++++++++++++++ parse file list +++++++++++++++++++++++++++++++++++++++
	for (i=0; i<ItemsInList(fileList); i++)
		filePath = folderPath + StringFromList(i,fileList)
		wave/Z/T logInfo = SpectraWatchDog#WD_readSpectraInfo(filePath)							// read info from the file => spits out a two-column text wave
		
		if (strlen(zipNamePrefix))																// cleanup
			DeleteFile/Z filePath
		endif
		
		if (!WaveExists(logInfo))																// not a supported file => abort
			error = -1
			continue
		endif
		InsertPoints 0,1, logInfo;
		logInfo[0][] = {{"File Name"},{ParseFilePath(0, filePath, ":", 1, 0)}}					// the first entry is always the file name in the log
		
		int panelCall = ParamIsDefault(logFile)
		if (panelCall)
			logFile = getLogPath()
		endif
		// +++++++++++++++++++++++++++++++ write the log file ++++++++++++++++++++++++++++++++++++
		string header = "", entries = ""
//		string entryID = "";	sPrintf entryID, "[%.4d] ", countLogFileEntries(logFile)
		logInfo[0][1] = zipNamePrefix + logInfo[0][1]											// add unique identifier to file name (in case of overlapping names)
		Duplicate/Free/T/RMD=[][0] logInfo, tempHeader
		Duplicate/Free/T/RMD=[][1] logInfo, tempEntries
		
		int logID = 0
		Open/Z logID as logFile
		if (V_Flag == 0)
			FStatus logID
			if (V_logEOF == 0)																	// is new file => add header	
				wfprintf header, "%s"+kDelimiterChar, tempHeader
				header = RemoveEnding(header, kDelimiterChar)+"\r\n"
			endif
			wfprintf entries, "%s"+kDelimiterChar, tempEntries
			entries = RemoveEnding(entries, kDelimiterChar)+"\r\n"
			FSetPos logID, V_logEOF
			FPrintF logID, header + entries
			Close logID
		else
			Print "Error: The log file is locked by another program and cannot be updated. Make sure the log file is not open at the same time, e.g., in Excel."
			error = -2
			break
		endif
		
		if (panelCall)																			// only for the watchdog panel => write the local log wave
			SpectraWatchDog#WD_addLogListEntry(logInfo)
		endif
	endfor
	return error
end

//################################################################################################

static function WD_addLogListEntry(wave/T logInfo)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logListNames	= stg:WD_logInfoFileNames
	SVAR colRepeat		= stg:WD_RepeatColumn
	wave/T logList		= stg:WD_logInfoDispList
	wave selList		= stg:WD_logInfoSelList
	
	int i, selSize = DimSize(selList,1), infoSize = DimSize(logInfo,0)-1
	if (DimSize(logList,1) < infoSize)
		Redimension/N=(-1, infoSize, -1) logList, selList
	endif
	
	if (strlen(logList[0][0]))																	// if first row already contains an entry
		Redimension/N=(DimSize(logList,0)+1,-1,-1) logList, selList
		for (i = 0; i < infoSize; i++)															// repeat entries if needed
			variable isRepeated = str2num(StringFromList(i,colRepeat,","))
			isRepeated = numtype(isRepeated) != 0 ? 0 : isRepeated
			if (isRepeated)
				logInfo[i+1][1] = logList[DimSize(logList,0)-2][i]								// use values from previous entry
			endif
		endfor
	else
		for (i = 0; i < infoSize; i++)															// write labels for the first time
			string title = cleanStr(logInfo[i+1][0],"\"|'|,|;|")[0,255]
			SetDimLabel 1,i,$title, logList
		endfor
	endif
	
	logList[DimSize(logList,0)-1][0,infoSize-1] = logInfo[q+1][1]
	selList[DimSize(selList,0)-1][][0] = 0x02
	if (DimSize(selList,1) > selSize)															// new column(s) added => make select-able
		selList[][selSize,][0] = 0x02
	endif
	SpectraWatchDog#WD_colorListWave(selList)
	logListNames += logInfo[0][1] + ";"
	return 0
end

static function/S cleanStr(string str, string list [, string insert])
	int i; string replace = SelectString(ParamIsDefault(insert),insert,"")
	for (i = 0; i < ItemsInList(list,"|"); i++)
		str = ReplaceString(StringFromList(i,list,"|"), str, replace)
	endfor
	return str
end

//################################################################################################

static function countLogFileEntries(string logFilePath)
	int i=0, fileID
	string read
	Open/Z/R fileID as logFilePath
	if (!V_flag)
		do
			FReadLine fileID, read; i++
		while (strlen(read) > 0)
		Close fileID
	endif
	return i > 0 ? i-1 : 1
end

//################################################################################################

function WD_SpectraWatchFolderBT(STRUCT WMBackgroundStruct &s)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	NVAR startCount	= stg:WD_startCounter
	NVAR runCount	= stg:WD_runCounter
	runCount = (s.curRunTicks-startCount)/60													// record passed time since start
	SpectraWatchDog#WD_checkNewFiles()
	return 0
end

//################################################################################################

static function WD_checkNewFiles()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR targetDir	= stg:WD_TargetFolder
	SVAR doneList	= stg:WD_ProcessedFileList
	SVAR targetExt	= stg:WD_TargeFileExt
	
	NewPath/O/Q WD_WatchFolderPath, targetDir													// make sure the path exists and is not overwritten elsewhere
	int i; string currFileList = ""
	for (i = 0; i < ItemsInList(targetExt); i++)
		currFileList += IndexedFile(WD_WatchFolderPath, -1, "."+StringFromList(i,targetExt))	// list of files in the folder to watch
	endfor
	string newFileList = RemoveFromList(doneList, currFileList)									// compare to already processed list
	
	for(i = 0; i < ItemsInList(newFileList); i++)
		string fileName = StringFromList(i, newFileList)
		if (SpectraWatchDog#FileIsNotLocked(targetDir+fileName,0))								// ensures that the file is closed before processing it to avoid processing while the other program is still writing it.
			SpectraWatchDog#WD_processFile(ParseFilePath(5,targetDir,":",0,0)+fileName, extList=targetExt)		// pass the file path as Mac formatted
			doneList += fileName+";"
		endif
	endfor
	return 0
end

//################################################################################################

static function FileIsNotLocked(string fullPath, variable doWrite)
	variable refNum = 0
	if (doWrite)
		Open/Z refNum as fullPath
	else
		Open/R/Z refNum as fullPath
	endif
	if (refNum != 0)																			// file successfully opened?
		Close refNum
		return 1
	endif
	return 0
end

//################################################################################################
//	GUI
//################################################################################################

static function buildWatchDogPanel()
	if (WinType("SpectraWatchDogPanel") == 7)
		DoWindow/F SpectraWatchDogPanel
		return 0
	endif
	
	string colWidths = StringByKey("Column_Widths", SpectraWatchDog#WD_readSettingsFromFile(), "=")
	if (!strlen(colWidths))
		colWidths = kDefaultColumnWidhts
	endif
	
	SpectraWatchDog#WD_prepareWorkingFolder()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	
	variable W = 765, H = 600
	variable extW = W + 560
	NewPanel/K=1/N=SpectraWatchDogPanel/W=(200,80,200+extW,80+H) as "Spectra Watchdog (ver. "+num2str(kVersion)+" - "+kVersionDate+")"
	variable px = PanelResolution("SpectraWatchDogPanel")/screenResolution
	SetWindow SpectraWatchDogPanel, sizelimit={W*px, H*px, inf, inf}
	
	Setvariable v_TargetFolder	,pos={11,8}		,size={W-240,23}	,bodywidth=W-340	,title="Folder to Watch:"		,help={"Sets the folder that the Watchdog watches."}
	Setvariable v_LogFileFolder	,pos={11,34}	,size={W-240,23}	,bodywidth=W-340	,title="Save Log File to:"		,help={"Sets the folder to which the output log is written to."}
	Button b_SelectTargetFolder	,pos={W-223,7}	,size={70,20}							,title="Select"
	Button b_SelectOutLogFolder	,pos={W-223,33}	,size={70,20}							,title="Select"
	Button b_ClearLogList		,pos={W-143,60}	,size={130,20}							,title="Clear Display List"		,help={"Clears the currently displayed list (NOT the log file). Already processed files will not be reprocessed."}
	Button b_ResetWatchDog		,pos={W-253,60}	,size={100,20}							,title="Reset All"				,help={"Completely resets watchdog to the initial state. All files will reprocessed after restart."}
	Button b_LoadFromLogFile	,pos={W-143,33}	,size={130,20}							,title="Read Log Entries"		,help={"Loads past log file entries into the current watchdog session."}
	Button b_StartStopButton	,pos={180,60}	,size={140,20}							,title="Start Watchdog"			,help={"Starts or stops the Watchdog task."}
	Button b_ReadOnceButton		,pos={330,60}	,size={140,20}							,title="Read Folder Once"		,help={"Just check the folder once without starting the background task."}
	Setvariable v_TargetFileExt	,pos={W-135,8}	,size={123,23}		,bodywidth=55		,title="File Type(s): *."		,help={"Sets the file type (ending) which the watchdog will look out for. Only supported file types should be entered here. Multiple extensions can be watched at the same time. For this, enter a semicolon-separated list."}
	CheckBox c_ReadOnlyNew		,pos={W,9}		,size={140,23}							,title="Ignore Existing Files"	,help={"Ignores already present files in the selected folder when starting the watchdog. Only newly added files will be processed."}
	Setvariable v_runCountDisp	,pos={11,61}	,size={160,23}		,bodywidth=60		,title="Running for:"

	ListBox WD_DataDisplayList	,pos={11,85}	,size={extW-22,H-96}
	ListBox WD_DataDisplayList	,listWave = stg:WD_logInfoDispList
	ListBox WD_DataDisplayList	,selWave  = stg:WD_logInfoSelList
	ListBox WD_DataDisplayList	,colorWave= stg:WD_logInfoColors
	ListBox WD_DataDisplayList	,proc=WD_ListHook ,frame=1 ,focusRing=0 ,mode=10 ,editStyle=1 ,userColumnResize=1 ,clickEventModifiers=5
	Execute/Q/Z "ListBox WD_DataDisplayList ,win=SpectraWatchDogPanel ,widths={"+colWidths+"}"
	
	Button b_SelectTargetFolder	,proc=WD_SelectInputFolderProc
	Button b_SelectOutLogFolder	,proc=WD_SelectLogFileFolderProc
	Button b_ClearLogList		,proc=WD_ClearAndResetButton
	Button b_ResetWatchDog		,proc=WD_ClearAndResetButton		,fstyle=2
	Button b_LoadFromLogFile	,proc=WD_ReadFromLogButton
	Button b_StartStopButton	,proc=WD_StartTaskButton			,fstyle=1	,valueColor=(0,16000,65000)
	Button b_ReadOnceButton		,proc=WD_StartTaskButton			,fstyle=1	,valueColor=(0,40000,16000)
	Setvariable v_TargetFolder	,value=stg:WD_TargetFolder	,proc=WD_SetLogFileFolderProc
	Setvariable v_TargetFileExt	,value=stg:WD_TargeFileExt	,proc=WD_SetLogFileFolderProc
	Setvariable v_LogFileFolder	,value=stg:WD_LogFilePath	,proc=WD_SetLogFileFolderProc
	Setvariable v_runCountDisp	,value=stg:WD_runCounter	,noedit=1	,limits={-inf,inf,0} ,format="%d s"
	
	SetWindow SpectraWatchDogPanel, hook(panelHook) = WD_PanelHook
	SpectraWatchDog#enableDisableControls()
	return 0
end

//################################################################################################

static function/S getLogPath()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logPath = stg:WD_LogFilePath
	return logPath
end

//################################################################################################

static function enableDisableControls()															// disables all controls other than start button and runtime counter
	string ctrlList = RemoveFromList("b_StartStopButton;v_runCountDisp;WD_DataDisplayList;",ControlNameList("SpectraWatchDogPanel",";","*"))
	CtrlNamedBackground SpectraWatchDogTask, status
	int isRunning = NumberByKey("RUN", S_info)
	ModifyControlList ctrlList	,win=SpectraWatchDogPanel ,disable=2*isRunning
	Button b_StartStopButton	,win=SpectraWatchDogPanel ,title=SelectString(isRunning,"Start Watchdog","Stop Watchdog")
end

//################################################################################################

static Constant kSpecialColor = 13																// row where special colors start

static function WD_prepareWorkingFolder()
	DFREF savDF = GetDataFolderDFR()
	NewDataFolder/O/S root:Packages
	NewDataFolder/O/S root:Packages:$(kWatchdogWorkingDir)
		string settings = SpectraWatchDog#WD_readSettingsFromFile()								// load previous settings from disc
		string colRepeat = StringByKey("Column_Repeat", settings, "=")
		string targetExt = StringByKey("Spectra_Extension", settings, "=")
		if (!strlen(colRepeat))
			colRepeat = kDefaultRepeatColumn
		endif
		if (!strlen(targetExt))
			targetExt = kDefaultExtensionStr
		endif
		String/G WD_TargetFolder = StringByKey("Target", settings, "=")
		String/G WD_LogFilePath  = StringByKey("LogDir", settings, "=")
		String/G WD_TargeFileExt = targetExt
		String/G WD_RepeatColumn = colRepeat

		if (!Exists("WD_runCounter"))
			Variable/G WD_runCounter = 0
		endif
		if (!Exists("WD_startCounter"))
			Variable/G WD_startCounter = 0
		endif
		if (!Exists("WD_ProcessedFileList"))
			String/G WD_ProcessedFileList = ""													// the list of files which have been read (logged or not)
		endif
		
		wave/Z/T list = WD_logInfoDispList														// the local log list
		if (!WaveExists(list))
			String/G WD_logInfoFileNames = ""													// NEEDS TO MATCH WD_logInfoDispList in size at all times => list of files which actually have been logged locally
			Make/T/N=(1,1) root:Packages:$(kWatchdogWorkingDir):WD_logInfoDispList/wave=list
		endif
		
		Make/O/N=(DimSize(list,0),DimSize(list,1),3) WD_logInfoSelList = 0
		SetDimLabel 2,1,backColors,WD_logInfoSelList											// define plane 1 as background colors
		SetDimLabel 2,2,foreColors,WD_logInfoSelList											// define plane 2 as foreground colors
		if (DimSize(list,1) > 1)
			WD_logInfoSelList[][][0] = 0x02														// only make editable if there are contents
		endif
		
		Make/W/U/O/N=(0,3) WD_logInfoColors = 65535												// define some colors
		WD_logInfoColors[0][]  = {{255},{255},{255}}	// white
		WD_logInfoColors[1][]  = {{255},{243},{243}}	// red
		WD_logInfoColors[2][]  = {{255},{243},{252}}	// pink
		WD_logInfoColors[3][]  = {{252},{243},{255}}	// purple
		WD_logInfoColors[4][]  = {{243},{243},{255}}	// blue
		WD_logInfoColors[5][]  = {{243},{255},{255}}	// cyan
		WD_logInfoColors[6][]  = {{243},{255},{243}}	// green
		WD_logInfoColors[7][]  = {{255},{255},{240}}	// yellow
		WD_logInfoColors[8][]  = {{255},{243},{233}}	// orange
		WD_logInfoColors[9][]  = {{243},{243},{243}}	// gray
		WD_logInfoColors[10][] = {{220},{240},{255}}	// dark blue
		WD_logInfoColors[11][] = {{220},{255},{220}}	// dark green
		WD_logInfoColors[12][] = {{255},{220},{220}}	// dark red
		WD_logInfoColors[13][] = {{110},{160},{255}}	// selection
		WD_logInfoColors *= 257
		
		if (DimSize(list,1) > 1)																// recolor columns after reopening the panel
			SpectraWatchDog#WD_colorListWave(WD_logInfoSelList)
		endif
	SetDataFolder savDF
	return 0
end

//################################################################################################

function WD_ListHook(STRUCT WMListboxAction &s) : ListboxControl
//	if (s.eventCode <3 || s.eventCode > 5)														// debug
//		Print s.eventCode, s.eventMod
//	endif
	
	int nRows, retValue = 0
	Switch (s.eventCode)
		case 1:		// mouse-down
			if (s.eventMod & 16)																// right-click menu
				retValue = 1
				if (s.row == -1)																// title row
					SVar colRepeat = GetWavesDataFolderDFR(s.listWave):WD_RepeatColumn
					variable isRepeated = str2num(StringFromList(s.col,colRepeat,","))
					PopupContextualMenu "white;red;pink;purple;blue;cyan;green;yellow;orange;gray;dark blue;dark green;dark red;-;Default Widths;Optimal Widths;Equal Widths;-;"+SelectString(isRepeated,"","!*")+"Repeat Values;"
					if (V_Flag > 0)
						if (V_Flag-1 < kSpecialColor)
							s.SelWave[][s.col][1] = V_Flag-1									// color column
						elseif (StringMatch(S_selection,"* Widths"))
							string colWidth = kDefaultColumnWidhts
							if (CmpStr(S_selection,"Optimal Widths") == 0)						// tries to find optimal column widths from the contents
								Make/Free/N=(DimSize(s.listWave,0),DimSize(s.listWave,1)) tempLen = strlen(s.listWave[p][q])
								MatrixOp/Free tempMat = maxCols(tempLen)^t
								MatrixOp/Free tempLen = replace(tempMat,0,5)/minVal(replace(tempMat,0,inf))
								tempLen = tempLen[p] < mean(tempLen)/2 ? round(mean(tempLen)/2) : round(tempLen[p])
								wfprintf colWidth, "%d,", tempLen;	colWidth = RemoveEnding(colWidth,",")
							elseif (CmpStr(S_selection,"Equal Widths") == 0)
								#if (IgorVersion() >= 9.0)
									colWidth = ReplicateString("1,", DimSize(s.SelWave,1)-1) + "1"
								#else
									colWidth = "1"
								#endif
							endif
							Execute/Q/Z "ListBox WD_DataDisplayList ,win="+s.win+" ,widths={"+colWidth+"}"
						elseif (CmpStr(S_selection,"Repeat Values") == 0)
							wave/T tempPara = ListToTextWave(colRepeat, ",")
							tempPara[s.col] = num2str(!isRepeated)								// toggle repeat setting
							wfprintf colRepeat, "%s,", tempPara
							colRepeat = RemoveEnding(colRepeat,",")
						endif
						SpectraWatchDog#WD_saveSettingsToFile()
					endif
				else
					nRows = DimSize(s.listWave, 0)
					PopupContextualMenu "Undo \ Redo;-;Cut;Copy;Paste;-;Select Whole Column;Select All Cells;-;Delete Selected Entries;"
					if (V_Flag > 0 && s.row < nRows)
						Switch (V_flag)
							case 1:
								SpectraWatchDog#doUndoRedo()
							break
							case 3:
							case 4:
							case 5:
								SpectraWatchDog#doCutCopyPaste(S_selection)
							break
							case 7:
								s.SelWave[][][0]	  = s.SelWave[p][q] & ~(2^0)				// first deselect all
								s.SelWave[][s.col][0] = s.SelWave[p][q] | (2^0)					// select column
							break
							case 8:
								s.SelWave[][][0]	  = s.SelWave[p][q] | (2^0)					// select all
							break
							case 10:
								s.listWave = SelectString((s.SelWave[p][q][0] & 2^0) != 0 || (s.SelWave[p][q][0] & 2^3) != 0, s.listWave[p][q], "")
								SpectraWatchDog#WD_updateLog()
							break
						EndSwitch
					endif
				endif
			endif
		break
		case 2:		// mouse up
			if (s.row == -1)																	// title row
				s.SelWave[][][0]	  = s.SelWave[p][q] & ~(2^0)								// first deselect all
				s.SelWave[][s.col][0] = s.SelWave[p][q] | (2^0)									// select column
			endif
			SpectraWatchDog#WD_colorListWave(s.SelWave)											// color the tag of all selected rows
			int i,j
			for (i = 0; i < DimSize(s.SelWave,0); i++)
				for (j = 0; j < DimSize(s.SelWave,1); j++)
					if ((s.SelWave[i][j][0] & 2^0) != 0 || (s.SelWave[i][j][0] & 2^3) != 0)
						s.SelWave[i][0][1] = kSpecialColor
					endif
				endfor
			endfor
		break
		case 6:		// begin edit
			SpectraWatchDog#WD_colorListWave(s.SelWave)
			s.SelWave[s.row][0][1] = kSpecialColor												// color tag
		break
		case 7:		// end edit
			SpectraWatchDog#WD_updateLog()
		break
		case 11:	// column resize
			SpectraWatchDog#WD_saveSettingsToFile()
		break
	EndSwitch
	return retValue
end

//################################################################################################

function WD_PanelHook(STRUCT WMWinHookStruct &s)
//	if (s.eventCode <3 || s.eventCode > 5)														// debug
//		Print s.eventCode, s.eventName, s.keycode, s.menuname, s.menuitem
//	endif

	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	
	variable hookTakeover = 0
	StrSwitch (s.eventName)
		case "menu":
			StrSwitch (s.menuitem)
				case "Copy":
					hookTakeover = 1
				case "Paste":
					SpectraWatchDog#doCutCopyPaste(s.menuitem)
				break
				case "Select All":
					//selList[][][0] = selList[p][q][r] | (2^0)
				break
			EndSwitch
		break
		case "earlyKeyboard":																	// only works in Igor 9
			Switch (s.keycode)
				case 65:	// ctrl + A
					selList[][][0] = selList[p][q][r] | (2^0)
				break
				case 88:	// ctrl + X
					SpectraWatchDog#doCutCopyPaste("Cut")
				break
				case 90:	// ctrl + Z
					SpectraWatchDog#doUndoRedo()
				break
				case 127:	// del key
					logList = SelectString((selList[p][q][0] & 2^0) != 0 || (selList[p][q][0] & 2^3) != 0, logList[p][q], "")
					SpectraWatchDog#WD_updateLog()
				break
				case 28:	// cursor keys
				case 29:
				case 30:
				case 31:
					Execute/Q/Z/P "SpectraWatchDog#WD_cursorColor()"							// needs to be executed after selection completes
				break
			EndSwitch
		case "keyboard":
		break;
		case "kill":
		break
		case "resize":
			ListBox WD_DataDisplayList ,win=$s.winName ,pos={11,85} ,size={s.winRect.right-22,s.winRect.bottom-96}
		break
	EndSwitch
	return hookTakeover
end

//################################################################################################

static function WD_cursorColor()																// dedicated function to color tags after keyboard cursor movement
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave selList = stg:WD_logInfoSelList
	SpectraWatchDog#WD_colorListWave(selList)
	int i,j
	for (i = 0; i < DimSize(selList,0); i++)
		for (j = 0; j < DimSize(selList,1); j++)
			if ((selList[i][j][0] & 2^0) != 0)
				selList[i][0][1] = kSpecialColor
			endif
		endfor
	endfor
	return 0
end

//################################################################################################

static function doUndoRedo()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T   logList = stg:WD_logInfoDispList
	wave/T/Z logBack = stg:WD_logInfoListBackup
	if (!WaveExists(logBack))
		return -1
	endif
	variable minRow = min(DimSize(logList,0),DimSize(logBack,0))-1
	variable minCol = min(DimSize(logList,1),DimSize(logBack,1))-1
	logList[,minRow][,minCol] = logBack[p][q]
	SpectraWatchDog#WD_updateLog()
	return 0
end

//################################################################################################

static function doCutCopyPaste(string which)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	
	int i, j, counter = 0, delete = 0
	StrSwitch (which)
		case "Cut":
			delete = 1
		case "Copy":
			string CopyStr = ""
			for (i = 0; i < DimSize(selList,0); i++)											// Excel-compatible copy function
				for (j = 0; j < DimSize(selList,1); j++)
					if ((selList[i][j][0] & 2^0) != 0 || (selList[i][j][0] & 2^3) != 0)			// if cell is selected append contents to copy-list
						if (counter == 0)														// find first selected row
							counter = i
						endif
						if(counter < i)															// step to next row has occurred
							CopyStr = RemoveEnding(CopyStr, "\t") + "\r\n"
							counter = i
						endif
						CopyStr += logList[i][j] + "\t"
						if (delete)
							logList[i][j] = ""
						endif
					endif
				endfor
			endfor
			CopyStr = RemoveEnding(CopyStr, "\t") + "\r\n"
			PutScrapText CopyStr
			if (delete)
				SpectraWatchDog#WD_updateLog()
			endif
		break
		case "Paste":
			string PasteStr = GetScrapText()													// paste directly into the list
			PasteStr = ReplaceString("\r", ReplaceString("\n", ReplaceString("\t", PasteStr, ";"), ";"), ";")
			PasteStr = RemoveFromList("", PasteStr)
			for (i = 0; i < DimSize(selList,0); i++)
				for (j = 0; j < DimSize(selList,1); j++)
					if ((selList[i][j][0] & 2^0) != 0 || (selList[i][j][0] & 2^3) != 0)
						logList[i][j] = StringFromList(counter,PasteStr)						// successively paste contents into selected cells
						counter += counter == ItemsInList(PasteStr)-1 ? 0 : 1					// if the end of the list is reached, repeat last element
					endif
				endfor
			endfor
			SpectraWatchDog#WD_updateLog()
		break
	EndSwitch
end

//################################################################################################

static function WD_colorListWave(wave selList)													// applies color to certain entries
	string colorList = StringByKey("Column_Colors", SpectraWatchDog#WD_readSettingsFromFile(), "=")
	if (!strlen(colorList))
		colorList = kDefaultColumnColors
	endif
	int i
	for (i = 0; i < DimSize(selList,1); i++)
		variable color = str2num(StringFromList(i,colorList,","))
		color = numtype(color) != 0 ? 0 : color
		selList[][i][1] = color
	endfor
	return 0
end

//################################################################################################

static function/S WD_readSettingsFromFile()
	int FileID
	string read = "", output = ""	
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFileName
	if (!V_flag)
		do
			FReadLine fileID, read
			output += read
		while (strlen(read) > 0)
		Close fileID
	endif
	output = ReplaceString("\r", output, ";")													// FReadLine terminates just with a line break
	return output
end

//################################################################################################

static function WD_saveSettingsToFile()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR TargetDir	 = stg:WD_TargetFolder
	SVAR logFilePath = stg:WD_LogFilePath
	wave/T logList	 = stg:WD_logInfoDispList
	wave selList	 = stg:WD_logInfoSelList
	SVar colRepeat	 = stg:WD_RepeatColumn
	SVar targetExt	 = stg:WD_TargeFileExt
	
	string colWidths = StringByKey("Column_Widths", SpectraWatchDog#WD_readSettingsFromFile(), "=")
	string colorList = StringByKey("Column_Colors", SpectraWatchDog#WD_readSettingsFromFile(), "=")
	if (!strlen(colWidths))
		colWidths = kDefaultColumnWidhts
	endif
	if (!strlen(colorList))
		colorList = kDefaultColumnColors
	endif
	
	if (DimSize(logList,1) > 1)
		Duplicate/Free/RMD=[0][][1] selList, tempColors											// save the current color settings
		if (tempColors[0] >= kSpecialColor)														// this color is reserved for highlighting
			Duplicate/Free/RMD=[][0][1] selList, tempFirstColor
			Extract/FREE/INDX tempFirstColor, colorIndex, tempFirstColor<kSpecialColor			// see if one of the rows is colored normally
			if (DimSize(colorIndex,0) > 0)
				tempColors[0] = tempFirstColor[colorIndex[0]]
			else
				tempColors[0] = str2num(StringFromList(0,colorList,","))						// don't save the color for now
			endif
		endif
		wfprintf colorList, "%d,", tempColors
		colorList = RemoveEnding(colorList,",")
		ControlInfo/W=SpectraWatchDogPanel WD_DataDisplayList									// save the current width settings
		colWidths = S_columnWidths
	endif
	
	string Savestring = ""
	Savestring += "Target=" + ParseFilePath(5, TargetDir, "*", 0, 0) + "\r\n"
	Savestring += "LogDir=" + ParseFilePath(5, logFilePath, "*", 0, 0) + "\r\n"
	Savestring += "Spectra_Extension=" + targetExt + "\r\n"
	Savestring += "Column_Widths=" + colWidths + "\r\n"
	Savestring += "Column_Repeat=" + colRepeat + "\r\n"
	Savestring += "Column_Colors=" + colorList + "\r\n"
	
	int fileID = 0
	Open/Z fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFileName
	if (fileID != 0)
		FPrintF fileID, SaveString																// write settings => has to match with READING CODE in WD_prepareWorkingFolder() etc.
		Close fileID
	endif
end

//################################################################################################

function WD_SelectInputFolderProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	PathInfo WD_WatchFolderPath																	// check path status
	if (!V_flag)
		NewPath/O/Q WD_WatchFolderPath, SpecialDirPath("Desktop", 0, 0, 0)						// preset folder to the Desktop
	endif
	PathInfo/S WD_WatchFolderPath																// set open path to Desktop or last folder
	NewPath/O/M="Select folder for the watchdog process to watch."/Q WD_WatchFolderPath			// open folder dialog
	if (V_Flag != 0)
		return -1
	endif
	
	PathInfo WD_WatchFolderPath																	// keep path in global variable as well
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVar TargetDir = stg:WD_TargetFolder
	TargetDir = ParseFilePath(5, S_path, "*", 0, 0)
	
	SpectraWatchDog#WD_saveSettingsToFile()
	return 0
end

//################################################################################################

function WD_SelectLogFileFolderProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logFilePath = stg:WD_LogFilePath
	int fileID
	Open/D/F="Log Files (*"+kLogFileEnding+"):"+kLogFileEnding+";"/M="Chose the log name and location." fileID as logFilePath
	if(strlen(S_fileName))
		logFilePath = ParseFilePath(5, S_fileName, "*", 0, 0)									// the log file path only exists in the global and not as PATH variable
		SpectraWatchDog#WD_saveSettingsToFile()
		SpectraWatchDog#WD_askForLoadLog()
	endif
	return 0
end

//################################################################################################

function WD_SetLogFileFolderProc(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		if (CmpStr(s.ctrlName, "v_LogFileFolder") == 0)
			SpectraWatchDog#WD_askForLoadLog()
		endif
		SpectraWatchDog#WD_saveSettingsToFile()
	endif
	return 0
end

//################################################################################################

static function WD_askForLoadLog()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T logList = stg:WD_logInfoDispList
	
	GetFileFolderInfo/Z/Q getLogPath()															// if the log file already exists
	if (V_flag == 0 && V_isFile == 1)
		DoAlert/T="Log File Already Exists" 1, "Selected log file already exists. Load previous log contents into current session?"
		if (V_Flag == 1)
			SpectraWatchDog#WD_ReadFromLog()
		endif
	else
		if (strlen(logList[0][0]))
			DoAlert/T="Initialize New Log File" 1, "A new log file will be created. Also clear entries from currently displayed list (saved log entries will NOT be deleted)?"
			if (V_Flag == 1)
				SpectraWatchDog#WD_ClearAndReset("ClearLogList")
			endif
		endif
	endif
	return 0
end

//################################################################################################

function WD_StartTaskButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR targetDir	= stg:WD_TargetFolder
	SVAR targetExt	= stg:WD_TargeFileExt
	SVAR logFilePath= stg:WD_LogFilePath
	SVAR doneList	= stg:WD_ProcessedFileList
	NVAR startCount	= stg:WD_startCounter
	NVAR runCount	= stg:WD_runCounter
	
	CtrlNamedBackground SpectraWatchDogTask, status
	if (NumberByKey("RUN", S_info))	// running
		CtrlNamedBackground SpectraWatchDogTask, stop
	else							// stopped
		targetDir = ParseFilePath(2, targetDir, SelectString(CmpStr(IgorInfo(2), "Macintosh")==0,"\\",":"), 0, 0)		// add ending separator, if not already there
		GetFileFolderInfo/Z/Q targetDir
		if (V_flag != 0 || V_isFolder != 1)
			DoAlert/T="Error: Folder not found." 0, "The selected target folder does not seem to be valid. Set a valid folder to watch first."
			return -1
		endif
		GetFileFolderInfo/Z/Q ParseFilePath(1, ParseFilePath(5, logFilePath, "\\", 0, 0), "\\", 1, 0)
		if (V_flag != 0 || V_isFolder != 1)
			DoAlert/T="Error: Folder not found." 0, "The folder for the log file does not seem to be valid. Set a valid folder to write the log file to."
			return -2
		endif
		
		NewPath/O/Q WD_WatchFolderPath, targetDir												// make sure the path is properly set (even if entered manually)
		ControlInfo/W=$s.win c_ReadOnlyNew														// if already present files should be ignored
		if (V_Value)
			int i; string currFileList = ""
			for (i = 0; i < ItemsInList(targetExt); i++)
				currFileList += IndexedFile(WD_WatchFolderPath, -1, "."+StringFromList(i,targetExt))	// read all files in the current folder into processed list to check them off
			endfor
			doneList = RemoveFromList(currFileList, doneList) + currFileList
		endif
		
		GetFileFolderInfo/Z/Q logFilePath														// if the log file already exists
		if (V_flag == 0 && V_isFile == 1)
			wave/T/Z logFileData = SpectraWatchDog#WD_loadCurrentLog()
			if (WaveExists(logFileData))
				string logFileNames = processLogNameList(logFileData, extractZip=1)
				doneList = RemoveFromList(logFileNames, doneList) + logFileNames
			endif
		endif
		
		if (CmpStr(s.ctrlName,"b_ReadOnceButton") == 0)											// just briefly check folder and don't start the task
			SpectraWatchDog#WD_checkNewFiles()
		else
			startCount = ticks; runCount = 0
			CtrlNamedBackground SpectraWatchDogTask, period=kCheckFolderPeriod, proc=WD_SpectraWatchFolderBT, dialogsOK=0
			CtrlNamedBackground SpectraWatchDogTask, start
		endif
	endif
	
	SpectraWatchDog#enableDisableControls()
	return 0
end

//################################################################################################

function WD_ReadFromLogButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		SpectraWatchDog#WD_ReadFromLog()
	endif
	return 0
end

//################################################################################################

function WD_ClearAndResetButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		SpectraWatchDog#WD_ClearAndReset(ReplaceString("b_",s.ctrlName,""))
	endif
	return 0
end

//################################################################################################

static function WD_ClearAndReset(string which)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR doneList  = stg:WD_ProcessedFileList
	SVAR logNames  = stg:WD_logInfoFileNames
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	NVAR timeCount = stg:WD_runCounter
	StrSwitch (which)
		case "ResetWatchDog":
			DoAlert/T="Reset to initial state" 1, "Reset the watchdog to it's initial state?\r\rDo this to reprocess all files into a new log file (already saved log files will NOT be deleted or altered). You might want to change the log file name afterwards."
			if (V_Flag != 1)
				break
			endif
			doneList  = ""
			timeCount = 0
		case "ClearLogList":
			Redimension/N=(1,1,-1) logList,selList
			logNames  = ""																		// make sure to clear both logNames and logList
			logList   = ""
			selList   = 0
		break
	EndSwitch
	return 0
end

//################################################################################################

static function WD_ReadFromLog()
	DFREF savDF = GetDataFolderDFR()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logFilePath 	= stg:WD_LogFilePath
	SVAR logListNames	= stg:WD_logInfoFileNames
	SVAR doneList		= stg:WD_ProcessedFileList
	wave/T logList		= stg:WD_logInfoDispList
	wave selList		= stg:WD_logInfoSelList
	
	wave/Z/T logFileData = SpectraWatchDog#WD_loadCurrentLog()
	if (!WaveExists(logFileData))
		DoAlert 0, "Log file does not exist (yet)."
		return -1
	endif
	
	string zipListNames = ""
	logListNames = processLogNameList(logFileData, extractZip=0)
	zipListNames = processLogNameList(logFileData, extractZip=1)
	doneList = RemoveFromList(zipListNames, doneList) + zipListNames
	
	Redimension/N=(DimSize(logFileData,0)-1,-1,-1) logList, selList								// update display list
	if (DimSize(logList,1) < DimSize(logFileData,1)-1)
		Redimension/N=(-1,DimSize(logFileData,1)-1,-1) logList, selList
	endif
	int i
	for (i = 0; i < min(DimSize(logFileData,1)-1,DimSize(logList,1)); i++)						// write labels from the table
		string title = cleanStr(logFileData[0][i+1],"\"|'|,|;|")[0,255]
		SetDimLabel 1,i,$title, logList
	endfor
	logList = logFileData[p+1][q+1]
	selList[][][0] = 0x02
	SpectraWatchDog#WD_colorListWave(selList)
	return 0
end

//################################################################################################

static function/S processLogNameList(wave/T logData [,int extractZip])
	int i, doZip = ParamIsDefault(extractZip) ? 0 : extractZip									// extract only the name of zip files if present => zip contents not relevant for file watcher
	Make/Free/T/N=(DimSize(logData,0)-1) fileList = logData[p+1][0]								// save loaded file names into processed and logged list
	Extract/FREE/INDX fileList, emptyList, !strlen(fileList)									// delete empty entries
	for (i=DimSize(emptyList,0)-1; i>-1; i--)
		DeletePoints/M=0 emptyList[i],  1, fileList
		DeletePoints/M=0 emptyList[i]+1,1, logData
	endfor
	if (doZip)
		string zipFileName = ""
		for (i=DimSize(fileList,0)-1; i>-1; i--)
			SplitString/E=("^\[(.*)\]") fileList[i], zipFileName								// extract zip file name
			fileList[i] = SelectString(strlen(zipFileName),fileList[i],zipFileName)
		endfor
	endif
	FindDuplicates/Free/RT=fileList_clean fileList												// remove duplicate entries
	string logListNames = ""; wfprintf logListNames, "%s;", fileList_clean
	return logListNames
end

//################################################################################################

static function WD_updateLog()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logFilePath 	= stg:WD_LogFilePath
	SVAR logListNames	= stg:WD_logInfoFileNames
	wave/T logList		= stg:WD_logInfoDispList
	
	wave/T/Z logFileData = SpectraWatchDog#WD_loadCurrentLog()
	if (!WaveExists(logFileData))
		return -1
	endif
	if (DimSize(logFileData,1)-1 != DimSize(logList,1))											// make sure the local and loaded lists have the same no. of columns
		DoAlert 0, "The log file has the wrong number of entries and cannot be updated. Check the file for additional or missing columns."
		return -2
	endif
	if (ItemsInList(logListNames) != DimSize(logList,0))
		DoAlert 0, "Somehow the local log table is broken and the log file cannot be updated. Try to re-read the log entries from the file to fix this."
		return -3
	endif
	if (!SpectraWatchDog#FileIsNotLocked(logFilePath,1))										// file successfully opened?
		DoAlert 0, "The log file is locked by another program and cannot be updated. Make sure the log file is not open at the same time, e.g., in Excel."
		return -4
	endif
	
	Make/FREE/T/N=(DimSize(logFileData,0)) tempLogNames = logFileData[p][0]
	Make/FREE/T/N=(DimSize(logFileData,0)-1,DimSize(logFileData,1)-1) tempLogData = logFileData[p+1][q+1]
	
	int i
	string logFileNames = ""
	wfprintf logFileNames, "%s;", tempLogNames													// a list of file names in the log file
	for (i = 0; i < ItemsInList(logListNames); i++)
		int listPos = WhichListItem(StringFromList(i, logListNames), logFileNames)				// find entry in the log and replace the contents
		if (listPos > -1 && listPos)
			logFileData[listPos][1,] = logList[i][q-1]
		endif
	endfor
	
	// to make the log compatible with excel for entries which contain the delimiter character (e.g., comma), add "" around such strings
	Make/Free/D/N=(DimSize(logFileData,0), DimSize(logFileData,1)) findDelim = strsearch(logFileData[p][q],kDelimiterChar,0)
	logFileData = SelectString((findDelim[p][q] > -1), logFileData[p][q], "\""+logFileData[p][q]+"\"")
	
	Save/O/A=0/J/M="\r\n"/DLIM=kDelimiterChar logFileData as logFilePath						// overwrite log file
	if (!EqualWaves(tempLogData, logList, 1))
		Duplicate/O tempLogData, stg:WD_logInfoListBackup										// backup for undo functionality
	endif
	return 0
end

//################################################################################################

static function/WAVE WD_loadCurrentLog()
	GetFileFolderInfo/Z/Q getLogPath()
	if (V_flag != 0 || V_isFile != 1)
		return $""
	endif
	
	DFREF savDF = GetDataFolderDFR()
	SetDataFolder NewFreeDataFolder()
		LoadWave/J/O/K=2/A/V={kDelimiterChar, "", 0, 0}/Q getLogPath()							// load log file
		variable success = V_flag
		if (success)
			Concatenate/Free/T WaveList("wave*",";","TEXT:1"), loadedData						// create one big table
		endif
	SetDataFolder savDF
	
	if (success)
		return loadedData
	else
		return $""
	endif
end